/*
 * Copyright (c) 2000-2014 Antonio De Lorenzo <antonio.delorenzo@develit.eu>
 *
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 *
 *   1.  Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *
 *   2.  Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MER-
 * CHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPE-
 * CIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTH-
 * ERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Alternatively, the contents of this file may be used under the terms of
 * the GNU General Public License ("GPL") version 2 or any later version,
 * in which case the provisions of the GPL are applicable instead of
 * the above. If you wish to allow the use of your version of this file
 * only under the terms of the GPL and not to allow others to use your
 * version of this file under the BSD license, indicate your decision
 * by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL. If you do not delete the
 * provisions above, a recipient may use your version of this file under
 * either the BSD or the GPL.
 */

#include "SmartThread.h"
#include "MessageLogger.h"
#include "Utility.h"

#include <stdlib.h>
#include <stdio.h>

#ifndef __WIN32__
	#include <unistd.h>
#else
	#include <windows.h>
#endif

#include <pthread.h>
#include <semaphore.h>

#include <string>

// SmartThread methods
SmartThread::SmartThread():
	m_running(false),
	m_stopPending(false),
	m_executionCounter(0){
}

SmartThread::~SmartThread(){
	Stop();
}

void SmartThread::SetRunning(const bool running){
	m_running=running;
	m_stopPending=false;
}

void SmartThread::IncExecutionCounter(){
	++m_executionCounter;
}

void SmartThread::Start(const unsigned int timeout){
	const unsigned int executionCounter(m_executionCounter);
	SmartThreadStart(*this);
	if(0<timeout){
		for(unsigned int i(0); i<(timeout/32) && !IsRunning(); ++i)
			SmartSleep(32);
			
		if(!IsRunning() && executionCounter==m_executionCounter){
			throw(LOGGER(0, LM_ERROR, "SmartThread::Start", "Unable to start thread"));
		}
	}
}

void SmartThread::Stop(const unsigned int timeout){
	SetStopPending();
	for(unsigned int i(0); i<(timeout/32) && IsRunning(); ++i)
		SmartSleep(32);
		
	if(IsRunning()){
		throw(LOGGER(0, LM_ERROR, "SmartThread::Stop", "Unable to stop thread"));
	}
}

bool SmartThread::IsRunning() const{
	return m_running;
}

void SmartThread::SetStopPending(){
	m_stopPending=true;
}

bool SmartThread::IsStopPending() const{
	return m_stopPending;
}

// SmartSemaphore methods
#define VOID_TO_HSEM(p) reinterpret_cast<sem_t*>(p)

SmartSemaphore::SmartSemaphore(const unsigned int initialValue):
	m_handle(new sem_t()){
	
	if(0!=sem_init(VOID_TO_HSEM(m_handle), 0, initialValue))
		throw(LOGGER(0, LM_ERROR, "SmartSemaphore::SmartSemaphore", "Unable to initialize semaphore"));
}

SmartSemaphore::~SmartSemaphore(){
	sem_destroy(VOID_TO_HSEM(m_handle));
	delete VOID_TO_HSEM(m_handle);
}
	
void SmartSemaphore::Post(){
	if(0!=sem_post(VOID_TO_HSEM(m_handle)))
		throw(LOGGER(0, LM_ERROR, "SmartSemaphore::SmartSemaphore", "Unable to post semaphore"));
}

void SmartSemaphore::Wait(){
	if(0!=sem_wait(VOID_TO_HSEM(m_handle)))
		throw(LOGGER(0, LM_ERROR, "SmartSemaphore::SmartSemaphore", "Unable to wait semaphore"));
}

// CriticalSection methods
CriticalSection::CriticalSection(SmartSemaphore& semaphore):
	m_semaphore(semaphore){
		
	m_semaphore.Wait();
}

CriticalSection::~CriticalSection(){
	m_semaphore.Post();
}

// SmartThreadStart
void* SmartThreadKicker(void* smartThread){
	SmartThread& threadInstance(*reinterpret_cast<SmartThread*>(smartThread));
	
	threadInstance.SetRunning(true);
		
	try{
		threadInstance.Run();
	}catch(...){
	}
	
	threadInstance.IncExecutionCounter();
	threadInstance.SetRunning(false);
	
	return NULL;
}

bool SmartThreadStart(SmartThread& smartThread){
	pthread_attr_t attr;
	pthread_attr_init(&attr);
	
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	
	pthread_t thread;
	const bool result(0==pthread_create(&thread, &attr, SmartThreadKicker, &smartThread));
	pthread_attr_destroy(&attr);

	return result;
}

// SmartSleep
bool SmartSleep(const unsigned int msec){
	#ifndef __WIN32__
		usleep(msec*1000);
	#else
		Sleep(msec);
	#endif
	
	return true;
}
