/*
 * Copyright (c) 2000-2014 Antonio De Lorenzo <antonio.delorenzo@develit.eu>
 *
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 *
 *   1.  Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *
 *   2.  Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MER-
 * CHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPE-
 * CIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTH-
 * ERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Alternatively, the contents of this file may be used under the terms of
 * the GNU General Public License ("GPL") version 2 or any later version,
 * in which case the provisions of the GPL are applicable instead of
 * the above. If you wish to allow the use of your version of this file
 * only under the terms of the GPL and not to allow others to use your
 * version of this file under the BSD license, indicate your decision
 * by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL. If you do not delete the
 * provisions above, a recipient may use your version of this file under
 * either the BSD or the GPL.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <Utility.h>
#include <MessageLogger.h>
#include <CircularBuffer.h>

#include <lzf.h>

#define COMPRESSED_FILE_EXTENSION ".xtc"

#define MEGABYTE 1048576
#define BUFFERS 32
#define BUFFERSIZE MEGABYTE

unsigned long GetFileSize(FILE* file){
	fseek(file, 0L, SEEK_END);
	const unsigned long fileSize(ftell(file));
	fseek(file, 0L, SEEK_SET);
	
	return fileSize;
}

// ReadThread
class ReadThread: public SmartThread{
private:
protected:
	int& m_result;
	FILE* m_inputFile;
	CircularBuffer& m_readBuffer;
	const unsigned long m_inputFileSize;
	unsigned long m_transferredBytes;
	
public:
	ReadThread(int& result, FILE* inputFile, CircularBuffer& readBuffer);
	virtual ~ReadThread();
	
	unsigned long GetTranferredBytes();
	int GetTranferredPercentage();
	
	virtual void Run();
};

ReadThread::ReadThread(int& result, FILE* inputFile, CircularBuffer& readBuffer):
	SmartThread(),
	m_result(result),
	m_inputFile(inputFile),
	m_readBuffer(readBuffer),
	m_inputFileSize(GetFileSize(m_inputFile)),
	m_transferredBytes(0){
		
	Start();
}

ReadThread::~ReadThread(){
	printf("info: read  %lu bytes\n", m_transferredBytes);
}

unsigned long ReadThread::GetTranferredBytes(){
	return m_transferredBytes;
}

int ReadThread::GetTranferredPercentage(){
	return 100L*m_transferredBytes/m_inputFileSize;
}
	
void ReadThread::Run(){
	char buffer[BUFFERSIZE];
	int bufferSize(0);
	DataBuffer* data(NULL);
	
	while(!IsStopPending()){
		if(0==bufferSize){
			bufferSize=fread(buffer, sizeof(char), sizeof(buffer), m_inputFile);
			if(0==bufferSize){
				break;
			}
			
			m_transferredBytes+=bufferSize;
		}
		
		data=m_readBuffer.GetFrontBuffer();
		if(NULL==data){
			SmartSleep(128);
		} else{
			data->SetBuffer(buffer, bufferSize);	
			m_readBuffer.IncFrontIndex();
			bufferSize=0;
		}
	}
	
	if(m_inputFileSize!=m_transferredBytes){
		m_result=1;
		printf("\n error: unable to read file\n");
	}
}

// ZThread
class ZThread: public SmartThread{
private:
protected:
	int& m_result;
	const bool m_compress;
	CircularBuffer& m_readBuffer;
	CircularBuffer& m_writeBuffer;
	SmartThread& m_readThread;
	unsigned long m_inputBytes;
	unsigned long m_outputBytes;
	
public:
	ZThread(int& result, const bool compress, CircularBuffer& readBuffer, CircularBuffer& writeBuffer, SmartThread& readThread);
	virtual ~ZThread();
	
	int GetCompressionRatio(){
		if(0==m_inputBytes || 0==m_outputBytes){
			return 0;
		}
		
		return 100-(m_compress ?(100L*m_outputBytes/m_inputBytes) :(100L*m_inputBytes/m_outputBytes));
	}
	
	virtual void Run();
};

ZThread::ZThread(int& result, const bool compress, CircularBuffer& readBuffer, CircularBuffer& writeBuffer, SmartThread& readThread):
	SmartThread(),
	m_result(result),
	m_compress(compress),
	m_readBuffer(readBuffer),
	m_writeBuffer(writeBuffer),
	m_readThread(readThread),
	m_inputBytes(0),
	m_outputBytes(0){
		
	Start();
}

ZThread::~ZThread(){
	printf("info: compression %d%% (in:%lu out:%lu)\n", GetCompressionRatio(), m_inputBytes, m_outputBytes);
}

void ZThread::Run(){
	DataBuffer incompletePacket;
	unsigned int bytesToCompletePacket(0);
	
	char zBuffer[BUFFERSIZE*2];
	unsigned int zBufferSize(0);
	DataBuffer* readData(NULL);
	DataBuffer* writeData(NULL);
	
	unsigned int bsize(0);
	
	while(!IsStopPending()){
		if(NULL==readData){
			readData=m_readBuffer.GetBackBuffer();
		}
		
		if(NULL==readData){
			if(!m_readThread.IsRunning()){
				break;
			}
			SmartSleep(128);
		} else {		
			writeData=m_writeBuffer.GetFrontBuffer();
			if(NULL==writeData){
				SmartSleep(128);
			} else{
				if(m_compress){
					bsize=readData->GetSize();
					zBufferSize=lzf_compress(readData->GetBuffer(), bsize, zBuffer, sizeof(zBuffer));
					writeData->SetBuffer(&zBufferSize, sizeof(zBufferSize));
					writeData->AppendBuffer(zBuffer, zBufferSize);
				} else{
					writeData->SetBuffer(NULL, 0);
					
					if(0!=bytesToCompletePacket){
						if(bytesToCompletePacket>readData->GetSizeLeft()){
							bsize=readData->GetSizeLeft();
							incompletePacket.AppendBuffer(readData->GetBuffer(), bsize);
							bytesToCompletePacket-=bsize;
						} else{
							incompletePacket.AppendBuffer(readData->GetBuffer(bytesToCompletePacket), bytesToCompletePacket);
							bsize=incompletePacket.GetSize();
							zBufferSize=lzf_decompress(incompletePacket.GetBuffer(), bsize, zBuffer, sizeof(zBuffer));
							writeData->AppendBuffer(zBuffer, zBufferSize);
							
							bytesToCompletePacket=0;
						}
					}
					
					if(0!=readData->GetSizeLeft()){
						unsigned int dataSize;
						const unsigned int packetSize(*static_cast<int*>(readData->GetBuffer(sizeof(zBufferSize), &dataSize)));
						if(packetSize>dataSize){
							incompletePacket.SetBuffer(readData->GetBuffer(), dataSize);
							bytesToCompletePacket=packetSize-dataSize;
						} else{
							zBufferSize=lzf_decompress(readData->GetBuffer(packetSize), packetSize, zBuffer, sizeof(zBuffer));
							writeData->AppendBuffer(zBuffer, zBufferSize);
						}
					}
				}
		
				if(0==readData->GetSizeLeft()){
					m_inputBytes+=readData->GetSize();
					m_readBuffer.IncBackIndex();
					readData=NULL;
				}
				
				m_outputBytes+=writeData->GetSize();
				m_writeBuffer.IncFrontIndex();
			}
		}
	}
}

// WriteThread
class WriteThread: public SmartThread{
private:
protected:
	int& m_result;
	FILE* m_outputFile;
	CircularBuffer& m_writeBuffer;
	SmartThread& m_zThread;
	unsigned long m_transferredBytes;
	
	bool CommitWriteBuffer(DataBuffer& buffer, const unsigned int limitToReach);
	
public:
	WriteThread(int& result, FILE* outputFile, CircularBuffer& writeBuffer, SmartThread& zThread);
	virtual ~WriteThread();
	
	virtual void Run();
	
	unsigned long GetTranferredBytes();
};

bool WriteThread::CommitWriteBuffer(DataBuffer& buffer, const unsigned int limitToReach){
	const unsigned int bsize(buffer.GetSize());
	if(bsize>limitToReach){
		if(1!=(fwrite(buffer.GetBuffer(), bsize, 1, m_outputFile))){
			m_result=1;
			printf("\nerror: unable to write the output file\n");
			return false;
		}
		buffer.SetBuffer(NULL, 0);
	}
	
	return true;
}

WriteThread::WriteThread(int& result, FILE* outputFile, CircularBuffer& writeBuffer, SmartThread& zThread):
	SmartThread(),
	m_result(result),
	m_outputFile(outputFile),
	m_writeBuffer(writeBuffer),
	m_zThread(zThread),
	m_transferredBytes(0){
		
	Start();
}

WriteThread::~WriteThread(){
	Stop();
	printf("info: wrote %lu bytes\n", m_transferredBytes);
}
	
void WriteThread::Run(){
	unsigned int bsize(0);
	
	DataBuffer writeBuffer;
	DataBuffer* data(NULL);
	while(!IsStopPending()){
		data=m_writeBuffer.GetBackBuffer();
		if(NULL==data){
			if(!m_zThread.IsRunning()){
				break;
			}
			SmartSleep(128);
		} else{
			bsize=data->GetSize();		
			writeBuffer.AppendBuffer(data->GetBuffer(), bsize);
			if(!CommitWriteBuffer(writeBuffer, BUFFERSIZE)){
				break;
			}
			
			m_transferredBytes+=data->GetSize();
			m_writeBuffer.IncBackIndex();
		}
	}
	
	CommitWriteBuffer(writeBuffer, 0);
}

unsigned long WriteThread::GetTranferredBytes(){
	return m_transferredBytes;
}

// main
int main(int argc, char** argv){
	if(3!=argc){
		printf("usage: %s <SrcFile> <DstFile>\n", argv[0]);
		return 1;
	}
	
	const bool showProgress(NULL!=strstr(argv[0], "progress"));
	
	const char* inputFileName(argv[1]);
	const char* outputFileName(argv[2]);
	
	const int ifnLen(strlen(inputFileName));
	if(0==ifnLen){
		printf("error: Input file name is empty");
		return 1;
	}
	const bool decompress(3<ifnLen && 0==strcmp(COMPRESSED_FILE_EXTENSION, inputFileName+ifnLen-4));	

	const int ofnLen(strlen(outputFileName));
	if(0==ofnLen){
		printf("error: Output file name is empty");
		return 1;
	}
	
	char ofnBuffer[512];
	if('/'==outputFileName[strlen(outputFileName)-1]){
		const char* ifn(inputFileName);
		for(const char* i(strchr(ifn, '/')); NULL!=i; ifn=i+1, i=strchr(ifn, '/'));
		 
		sprintf(ofnBuffer, "%s%s", outputFileName, ifn);
		if(decompress){
			ofnBuffer[strlen(ofnBuffer)-4]='\0';
		}
	} else{
		strcpy(ofnBuffer, outputFileName);
	}
	
	if(!decompress){
		strcat(ofnBuffer, COMPRESSED_FILE_EXTENSION);
	}
	
	outputFileName=ofnBuffer;
	
	printf("%s %s to %s\n", decompress ?"Decompressing" : "Compressing", inputFileName, outputFileName);
	
	int result(0);
	FILE* inputFile(fopen(inputFileName, "r"));
	FILE* outputFile(fopen(outputFileName, "w"));
	
	if(NULL==inputFile){
		printf("error: Unable to open %s for reading", inputFileName);
		result=1;
	}
	
	if(NULL==outputFile){
		printf("error: Unable to open %s for writing", outputFileName);
		result=1;
	}
	
	if(0==result){
		try{
			MessageLogger logger("/dev/null", LM_DEBUG, 2048, 4);
			MessageLoggerSingleton loggerSingleton(logger);
			
			unsigned long w0(0);
			unsigned long w1(0);
			time_t t0(time(NULL));
			time_t t1(t0);
			time_t startTime(t0);
			
			{
				CircularBuffer readBuffer(BUFFERS);
				CircularBuffer writeBuffer(BUFFERS);
				ReadThread readThread(result, inputFile, readBuffer);
				ZThread zThread(result, !decompress, readBuffer, writeBuffer, readThread); 
				WriteThread writeThread(result, outputFile, writeBuffer, zThread);
				
				while(writeThread.IsRunning()){
					t1=time(NULL);
					if(t1!=t0){
						w1=writeThread.GetTranferredBytes();
						
						if(showProgress){
							printf("Read %lu MB (%d%%) Wrote %lu MB (%lu MB/sec, C %d%%) - RQ=%d - WQ=%d         \r", 
								readThread.GetTranferredBytes()/MEGABYTE,
								readThread.GetTranferredPercentage(),
								w1/MEGABYTE,
								(w1-w0)/MEGABYTE,
								zThread.GetCompressionRatio(),
								readBuffer.GetBufferUsage(),
								writeBuffer.GetBufferUsage());
							fflush(stdout);
						}
						
						t0=t1;
						w0=w1;
					}
					
					SmartSleep(128);
				}
				
				printf("\n"); 
			}
			
			printf("Flushing output file...                                            \n");
			fflush(outputFile);
			fsync(fileno(outputFile));
			t1=time(NULL);
			printf("Copy time %ld seconds\n", t1-startTime);
			
		} catch(...){
			printf("error: Execption catched in main body");
		}
	}
	
	if(NULL!=inputFile){
		fclose(inputFile);
	}
	
	if(NULL!=outputFile){
		fclose(outputFile);
	}
	
	printf("\n");
	return result;
}
